rm(list=ls())

library("ProcData")
# install.packages("corrplot", dependencies = TRUE)
library(corrplot)

##########################################################################
### code chunk number 1: read and write process data from/to CSV files ###
##########################################################################
### This code chuck demonstrates functions read.seqs and write.seqs.   ###
### In both functions:                                                 ###
###    style specifies input/output file style                         ###
###    id_var specifies the column name of ids                         ###
###    action_var specifies the column name of action sequences        ###
###    time_var specifies the column name of time sequences            ###
###    step_sep is used only when style = "single". It specifies the   ###
###        separator of time steps                                     ###
##########################################################################

# read in multiple style
seqs_multiple <- read.seqs("example_multiple.csv", 
                           style = "multiple", id_var = "ID", 
                           action_var = "Action", time_var = "Time")
                           
# read in single style
seqs_single <- read.seqs("example_single.csv", style = "single", 
                         id_var = "ID", action_var = "Action", 
                         time_var = "Time", step_sep = ",")

all.equal(seqs_multiple, seqs_single)

# write in multiple style
write.seqs(seqs_multiple, file="seqs_format_multiple.csv", 
           style = "multiple", id_var="ID", action_var="Action", 
           time_var="Time", row.names=FALSE)
           
# write in single style
write.seqs(seqs_single, file="seqs_format_single.csv", 
           style = "single", id_var="ID", action_var="Action", 
           time_var="Time", step_sep=",", row.names=FALSE)
           
#############################################################################
###                 code chunk number 2: cc_data dataset                  ###
#############################################################################
### This code chunk demonstrates the climate control dataset in ProcData. ###
#############################################################################

# load cc_data
data("cc_data")
# cc_data is a list of two elements
names(cc_data)

# element seqs is an object of class "proc". It contains the response
# processes of 16,763 participants
class(cc_data$seqs)

# element responses is a binary vector containing the item responses
# 1: correct 0: incorrect
length(cc_data$responses)

table(cc_data$responses)

########################################################################
###                code chunk number 3: class "proc"                 ###
########################################################################
### This code chunk demonstrates S3 class "proc" for processes data. ###
########################################################################

# An object of class "proc" is a list of two elements.           
#    action_seqs: a list of character vectors (action sequences) 
#    time_seqs: a list of numeric vectors (timestamp sequences)

class(cc_data$seqs)
names(cc_data$seqs)

# print method of class "proc". 
# print.proc prints first 5 processes by default.
cc_data$seqs
# print first n processes
print(cc_data$seqs, n=5)

# you can specify the indices of processes to print.
print(cc_data$seqs, index=3)
print(cc_data$seqs, index=c(4,5))

# summary method of class "proc"
cc_summary <- summary(cc_data$seqs)
# It returns a list of summary statistics of processes
names(cc_summary)

# number of processes
cc_summary$n_seq

# number of unique actions
cc_summary$n_action 

# action set
sort(cc_summary$actions) # action set

#  process length
range(cc_summary$seq_length) # range of sequence lengths

# action frequency
sort(cc_summary$action_freq / sum(cc_summary$action_freq), decreasing=TRUE)[1:10]

# action sequence frequency
sort(cc_summary$action_seqfreq / cc_summary$n_seq, decreasing = TRUE)[1:10]

# action transition probability matrix
trans_mat <- cc_summary$trans_count
dim(trans_mat)
trans_mat <- trans_mat / rowSums(trans_mat) 
# randomly select 25 actions to plot the transition matrix
idx_plot <- sample(1:cc_summary$n_action, 25)
corrplot(trans_mat[idx_plot, idx_plot], 
		 is.corr = FALSE, cl.lim = c(0, 1), 
		 tl.cex = 0.75, na.label = " ")

# summary of total time
cc_summary$total_time

# summary of mean reaction time
cc_summary$mean_react_time


########################################################################
###              code chunk number 4: data processing                ###
########################################################################
### This part demonstrates five functions for data processing.       ###
########################################################################

# subset "proc" objects
seqs <- sub_seqs(cc_data$seqs, 4*1:10)
seqs
print(seqs, 1)

# remove repeated actions
seqs1 <- remove_repeat(seqs)
print(seqs1, 1)

# remove actions
seqs2 <- remove_action(seqs1, "0_0_0")
print(seqs2, 1)

# replace actions
seqs3 <- replace_action(seqs2, "reset", "RESET")
print(seqs3, 1)

# combine actions
seqs4 <- combine_actions(seqs2, c("0_0_1", "0_0_2"), "BOTTOM_MOVE_ONE")
print(seqs4, 1)


# Now let's process climate control data for later use.
# subset dataset
set.seed(1234)
n <- 1000
idx <- sample(1:cc_summary$n_seq, n)
cc_seqs <- sub_seqs(cc_data$seqs, idx)
y <- cc_data$responses[idx]

# check proportion of correct answers
table(y)
mean(y)

# remove repeated actions in action sequences
cc_seqs <- remove_repeat(cc_seqs)


########################################################################
### code chunk number 5: multidimensional scaling feature extraction ###
########################################################################
### The code in this chunk demonstrates the basic usage of functions ###
### seq2feature_mds and chooseK_mds for extracting mds features.     ###
########################################################################

# extract 10 features from the provided response processes
mds_res0 <- seq2feature_mds(seqs = cc_seqs, K = 5)
names(mds_res0)
dim(mds_res0$theta)

# select feature dimension K by cross-validation
mds_cv_res <- chooseK_mds(seqs = cc_seqs, K_cand = c(20, 50), n_fold = 3,
						  return_dist = TRUE)
# chooseK_mds returns a list containing the selected K, candidates of K, 
# cross validation loss for each candidate, and a dissimilarity matrix 
# if return_dist=TRUE.
names(mds_cv_res)
mds_cv_res$K
dim(mds_cv_res$dist_mat)
mds_cv_res$dist_mat[1:5, 1:5]

# pass the computed dissimilarity matrix to seq2feature_mds
mds_res <- seq2feature_mds(seqs = mds_cv_res$dist_mat, K = mds_cv_res$K)
plot(mds_res$theta[,1:2], xlab="feature 1", ylab="feature 2", pch=16, cex=0.7)



## use extracted features to predict item responses

# split the data into training, validation, and test sets
n_train <- n*0.8
n_test <- n - n_train
index_train <- sample(1:n, n_train)
index_test <- setdiff(1:n, index_train)

# fit a logistic regression model using the training set
mds_data <- data.frame(y = y, x = mds_res$theta)
mds_glm_res <- glm(y ~ ., family = "binomial", 
                   subset=index_train, data=mds_data)
# predict item responses from the fitted model on the test
yhat_test <- as.numeric(predict(mds_glm_res, newdata = mds_data[index_test,]) > 0)
# out-of-sample prediction accuracy
mean(y[index_test] == yhat_test)

summary(mds_glm_res)

# predict response outcome using only the first feature
mds_glm_res1 <- glm(y ~ x.1, family = "binomial", 
                    subset=index_train, data=mds_data)
yhat_test1 <- as.numeric(predict(mds_glm_res1, newdata = mds_data[index_test,]) > 0)
mean(y[index_test] == yhat_test1)


# Let's take a look at the interpretation of the first feature
# We will examine action sequences corresponding to the highest 
# and the lowest feature values

# First, sort the first feature from different respondents in ascending order
o_mds1 <- order(mds_res$theta[,1])
# print response processes corresponding to the lowest feature values
print(cc_seqs, index = head(o_mds1))
# print response processes corresponding to the highest feature values
print(cc_seqs, index = tail(o_mds1))

# plot values of the first feature againt log sequence length
seq_length <- sapply(cc_seqs$action_seqs, length)
plot(mds_res$theta[,1], log10(seq_length),
     xlab = "First MDS Feature", ylab = "log (process length)", pch = 16, cex=0.7)
# processes with the lowest values use vary one bar at a time strategy


########################################################################
###       code chunk number 6: autoencoder feature extraction        ###
########################################################################
### We demomstrate the basic usage of seq2feature_seq2seq and        ###
### chooseK_seq2seq in this code chunk.                              ###
########################################################################

# Training autoencoder takes longer time. A smaller simulated dataset is
# used for demonstration

# generate processes
seqs <- seq_gen(50, include_time=TRUE)

# consider action sequences only first
# select feature dimension by cross-validation
cv_res <- chooseK_seq2seq(seqs = seqs, ae_type = "action", 
                          K_cand = c(3, 5), rnn_type="gru", 
                          n_fold = 2, n_epoch = 5)
names(cv_res)
cv_res$K

# extract K action features
index_train <- sample(1:50, 40)
index_valid <- setdiff(1:50, index_train)
seq2seq_action_res <- seq2feature_seq2seq(seqs = seqs, ae_type = "action", 
                                   K = cv_res$K, rnn_type="gru", n_epoch=5,
                                   samples_train = index_train, 
                                   samples_valid = index_valid, 
                                   verbose = TRUE)
# seq2feature_seq2seq returns extracted features and the trace of loss function
# on the training and validation set.
names(seq2seq_action_res)
dim(seq2seq_action_res$theta)

# plot the loss trace for examining convergence.
plot(seq2seq_action_res$valid_loss, type="o", ylab="Loss on validation set", 
                                              xlab="Epoch")


# action-time sequence autoencoder features
# We can incorporate the time sequences by setting ae_type = "both".
seq2seq_at_res <- seq2feature_seq2seq(seqs = seqs, ae_type = "both", 
                              K = cv_res$K, rnn_type = "gru", n_epoch=5,
                              samples_train = index_train, 
                              samples_valid = index_valid, 
                              weights=c(0.5, 0.5),
                              verbose = TRUE)
names(seq2seq_at_res)
dim(seq2seq_at_res$theta)


########################################################################
###               code chunk number 7: sequence models               ###
########################################################################
### We demonstrate how to use seqm and its predict method to fit     ###
### sequence models and make predictions from the fitted model       ###
########################################################################

# subset cc_data
set.seed(1234)
n <- 200
samples <- sample(1:length(cc_data$responses), n)
seqs <- sub_seqs(cc_data$seqs, samples)
y <- cc_data$responses[samples]

# generate some covariates
x <- matrix(rnorm(n*2), ncol=2)

# split data
index_test <- sample(1:n, 0.2*n)
index_train <- setdiff(1:n, index_test)
seqs_train <- sub_seqs(seqs, index_train)
seqs_test <- sub_seqs(seqs, index_test)


## no covariate is used
res1 <- seqm(seqs = seqs_train, response = y[index_train], 
             response_type = "binary", actions=cc_summary$actions, 
             K_emb = 5, K_rnn = 5, n_epoch = 10, step_size=0.01)
class(res1)
names(res1)
res1$structure

# plot loss trace to check convergence
plot(res1$history[,1], ylim=range(res1$history), type="o", 
     ylab="Loss", xlab="Epoch", col="blue", pch=16)
lines(res1$history[,2], type="o", col="red", pch=4)
legend("topright", legend=c("train", "validation"), 
       col=c("blue", "red"), pch=c(16, 4), lty=1)

# predict method of seqm
pred_res1 <- predict(res1, new_seqs = seqs_test)
mean(as.numeric(pred_res1 > 0.5) == y[index_test])
mean(y[index_test])

# extract features 
theta1_seqm <- predict(res1, new_seqs = seqs, type="feature")
dim(theta1_seqm)

## use more hidden layers
res2 <- seqm(seqs = seqs_train, response = y[index_train],
             response_type = "binary", actions=cc_summary$actions, 
             K_emb = 5, K_rnn = 5, n_hidden=2, K_hidden=c(10,5), 
             n_epoch = 10, step_size=0.01)
class(res2)
res2$structure

pred_res2 <- predict(res2, new_seqs = seqs_test)
mean(as.numeric(pred_res2 > 0.5) == y[index_test])

## add covariates
res3 <- seqm(seqs = seqs_train, response = y[index_train],
             covariates = x[index_train, ],
             response_type = "binary", actions=cc_summary$actions,
             K_emb = 5, K_rnn = 5, n_epoch = 10, step_size=0.01)
pred_res3 <- predict(res3, new_seqs = seqs_test,
                     new_covariates=x[index_test, ])
mean(as.numeric(pred_res3 > 0.5) == y[index_test])

## include time sequences
res4 <- seqm(seqs = seqs_train, response = y[index_train],
             response_type = "binary", actions=cc_summary$actions,
             include_time=TRUE, time_interval=TRUE, log_time=TRUE,
             K_emb=5, K_rnn=5, n_epoch=10, step_size=0.01)
res4$structure

pred_res4 <- predict(res4, new_seqs = seqs_test)
mean(as.numeric(pred_res4 > 0.5) == y[index_test])


